# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles all S3 related configurations"""

import attr

try:
    import boto3
    from botocore.client import BaseClient
    from s3transfer.manager import TransferManager
except ImportError:
    print(
        "Missing libraries to support S3 functionality. Please re-install spock with the extra s3 dependencies -- "
        "pip install spock-config[s3]"
    )
from typing import Optional

# Iterate through the allowed download args for S3 and map into optional attr.ib
download_attrs = {
    val: attr.ib(
        default=None,
        type=str,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    for val in TransferManager.ALLOWED_DOWNLOAD_ARGS
}


# Make the class dynamically
S3DownloadConfig = attr.make_class(
    name="S3DownloadConfig", attrs=download_attrs, kw_only=True, frozen=True
)

# Iterate through the allowed upload args for S3 and map into optional attr.ib
upload_attrs = {
    val: attr.ib(
        default=None,
        type=str,
        validator=attr.validators.optional(attr.validators.instance_of(str)),
    )
    for val in TransferManager.ALLOWED_UPLOAD_ARGS
}


# Make the class dynamically
S3UploadConfig = attr.make_class(
    name="S3UploadConfig", attrs=upload_attrs, kw_only=True, frozen=True
)


@attr.s(auto_attribs=True)
class S3Config:
    """Configuration class for S3 support

    *Attributes*:

        session: instantiated boto3 session object
        s3_session: automatically generated s3 client from the boto3 session if not provided
        kms_arn: AWS KMS key ARN (optional)
        temp_folder: temporary working folder to write/read spock configuration(s) (optional: defaults to /tmp)
        download_config: S3DownloadConfig for extra download configs (optional)
        upload_config: S3UploadConfig for extra upload configs (optional)

    """

    session: boto3.Session
    # s3_session: BaseClient = attr.ib(init=False)
    s3_session: Optional[BaseClient] = None
    temp_folder: Optional[str] = "/tmp/"
    download_config: S3DownloadConfig = S3DownloadConfig()
    upload_config: S3UploadConfig = S3UploadConfig()

    def __attrs_post_init__(self):
        if self.s3_session is None:
            self.s3_session = self.session.client("s3")



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the optuna backend"""


import optuna

from spock.addons.tune.config import OptunaTunerConfig
from spock.addons.tune.interface import BaseInterface

try:
    from typing import TypedDict
except ImportError:
    from mypy_extensions import TypedDict


class OptunaTunerStatus(TypedDict):
    """Tuner status return object for Optuna -- supports the define-and-run style interface from Optuna

    *Attributes*:

        trial: current ask trial sample
        study: current optuna study object

    """

    trial: optuna.Trial
    study: optuna.Study


class OptunaInterface(BaseInterface):
    """Specific override to support the optuna backend -- supports the define-and-run style interface from Optuna

    *Attributes*:

        _map_type: dictionary that maps class names and types to fns that create optuna distributions
        _trial: current trial object from the optuna backend
        _tuner_obj: underlying optuna study object
        _tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types
        _param_obj: underlying object that optuna study can sample from (flat dictionary)
        _sample_hash: hash of the most recent sample draw

    """

    def __init__(self, tuner_config: OptunaTunerConfig, tuner_namespace):
        """OptunaInterface init call that maps variables, creates a map to fnc calls, and constructs the necessary
        underlying objects

        *Args*:

            tuner_config: configuration object for the optuna backend
            tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types

        """
        super(OptunaInterface, self).__init__(tuner_config, tuner_namespace)
        self._tuner_obj = optuna.create_study(
            **self._config_to_dict(self._tuner_config)
        )
        # Some variables to use later
        self._trial = None
        self._sample_hash = None
        # Mapping spock underlying classes to optuna distributions (define-and-run interface)
        self._map_type = {
            "RangeHyperParameter": {
                "int": self._uniform_int_dist,
                "float": self._uniform_float_dist,
            },
            "ChoiceHyperParameter": {
                "int": self._categorical_dist,
                "float": self._categorical_dist,
                "str": self._categorical_dist,
                "bool": self._categorical_dist,
            },
        }
        # Build the correct underlying dictionary object for Optuna
        self._param_obj = self._construct()

    @property
    def tuner_status(self) -> OptunaTunerStatus:
        return OptunaTunerStatus(trial=self._trial, study=self._tuner_obj)

    @property
    def best(self):
        rollup_dict, _ = self._sample_rollup(self._tuner_obj.best_trial.params)
        return (
            self._gen_spockspace(rollup_dict),
            self._tuner_obj.best_value,
        )

    @property
    def _get_sample(self):
        return self._tuner_obj.ask(self._param_obj)

    def sample(self):
        self._trial = self._get_sample
        # Roll this back out into a Spockspace so it can be merged into the fixed parameter Spockspace
        # Also need to un-dot the param names to rebuild the nested structure
        rollup_dict, sample_hash = self._sample_rollup(self._trial.params)
        self._sample_hash = sample_hash
        return self._gen_spockspace(rollup_dict)

    def _construct(self):
        optuna_dict = {}
        # These will only be nested one level deep given the tuner syntax
        for k, v in vars(self._tuner_namespace).items():
            for ik, iv in vars(v).items():
                param_fn = self._map_type[type(iv).__name__][iv.type]
                optuna_dict.update({f"{k}.{ik}": param_fn(iv)})
        return optuna_dict

    def _uniform_float_dist(self, val):
        """Assemble the optuna.distributions.(Log)UniformDistribution object

        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.UniformDistribution.html
        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.LogUniformDistribution.html

        *Args*:

            val: current attr val

        *Returns*:

            optuna.distributions.UniformDistribution or optuna.distributions.LogUniformDistribution

        """
        low, high = self._try_range_cast(val, type_string="RangeHyperParameter")
        log_scale = val.log_scale
        return (
            optuna.distributions.LogUniformDistribution(low=low, high=high)
            if log_scale
            else optuna.distributions.UniformDistribution(low=low, high=high)
        )

    def _uniform_int_dist(self, val):
        """Assemble the optuna.distributions.Int(Log)UniformDistribution object

        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.IntUniformDistribution.html
        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.IntLogUniformDistribution.html

        *Args*:

            val: current attr val

        *Returns*:

            optuna.distributions.IntUniformDistribution or optuna.distributions.IntLogUniformDistribution

        """
        low, high = self._try_range_cast(val, type_string="RangeHyperParameter")
        log_scale = val.log_scale
        return (
            optuna.distributions.IntLogUniformDistribution(low=low, high=high)
            if log_scale
            else optuna.distributions.IntUniformDistribution(low=low, high=high)
        )

    def _categorical_dist(self, val):
        """Assemble the optuna.distributions.CategoricalDistribution object

        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.CategoricalDistribution.html

        *Args*:

            val: current attr val

        *Returns*:

            optuna.distributions.CategoricalDistribution

        """
        val = self._try_choice_cast(val, type_string="ChoiceHyperParameter")
        return optuna.distributions.CategoricalDistribution(choices=val.choices)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles Spock data type wrappers"""

import argparse

import yaml


class Spockspace(argparse.Namespace):
    """Inherits from Namespace to implement a pretty print on the obj

    Overwrites the __repr__ method with a pretty version of printing

    """

    def __init__(self, **kwargs):
        super(Spockspace, self).__init__(**kwargs)

    def __repr__(self):
        # Remove aliases in YAML print
        yaml.Dumper.ignore_aliases = lambda *args: True
        return yaml.dump(self.__dict__, default_flow_style=False)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

import argparse
import sys
import typing
from pathlib import Path
from uuid import uuid4

import attr

from spock.backend.builder import AttrBuilder
from spock.backend.payload import AttrPayload
from spock.backend.saver import AttrSaver
from spock.backend.wrappers import Spockspace
from spock.utils import check_payload_overwrite, deep_payload_update


class ConfigArgBuilder:
    """Automatically generates dataclass instances from config file(s)

    This class builds out necessary arguments from *args classes, reads
    the arguments from specified config file(s), and subsequently (via chained
    call to generate) generates each class instance based on the necessary
    field values for each backend class instance

    *Attributes*:

        _args: all command line args
        _arg_namespace: generated argument namespace
        _builder_obj: instance of a BaseBuilder class
        _dict_args: dictionary args from the command line
        _payload_obj: instance of a BasePayload class
        _saver_obj: instance of a BaseSaver class
        _tune_payload_obj: payload for tuner related objects -- instance of TunerPayload class
        _tune_obj: instance of TunerBuilder class
        _tuner_interface: interface that handles the underlying library for sampling -- instance of TunerInterface
        _tuner_state: current state of the hyper-parameter sampler
        _tune_namespace: namespace that hold the generated tuner related parameters
        _sample_count: current call to the sample function
        _fixed_uuid: fixed uuid to write the best file to the same path

    """

    def __init__(
        self,
        *args,
        configs: typing.Optional[typing.List] = None,
        desc: str = "",
        no_cmd_line: bool = False,
        s3_config=None,
        **kwargs,
    ):
        """Init call for ConfigArgBuilder

        *Args*:

            *args: tuple of spock decorated classes to process
            configs: list of config paths
            desc: description for help
            no_cmd_line: turn off cmd line args
            s3_config: s3Config object for S3 support
            **kwargs: keyword args

        """
        # Do some verification first
        self._verify_attr(args)
        self._configs = configs
        self._no_cmd_line = no_cmd_line
        self._desc = desc
        # Build the payload and saver objects
        self._payload_obj = AttrPayload(s3_config=s3_config)
        self._saver_obj = AttrSaver(s3_config=s3_config)
        # Split the fixed parameters from the tuneable ones (if present)
        fixed_args, tune_args = self._strip_tune_parameters(args)
        # The fixed parameter builder
        self._builder_obj = AttrBuilder(*fixed_args, **kwargs)
        # The possible tunable parameter builder -- might return None
        self._tune_obj, self._tune_payload_obj = self._handle_tuner_objects(
            tune_args, s3_config, kwargs
        )
        self._tuner_interface = None
        self._tuner_state = None
        self._tuner_status = None
        self._sample_count = 0
        self._fixed_uuid = str(uuid4())
        try:
            # Get all cmd line args and build overrides
            self._args = self._handle_cmd_line()
            # Get the actual payload from the config files -- fixed configs
            self._dict_args = self._get_payload(
                payload_obj=self._payload_obj,
                input_classes=self._builder_obj.input_classes,
                ignore_args=tune_args,
            )
            # Build the Spockspace from the payload and the classes
            # Fixed configs
            self._arg_namespace = self._builder_obj.generate(self._dict_args)
            # Get the payload from the config files -- hyper-parameters -- only if the obj is not None
            if self._tune_obj is not None:
                self._tune_args = self._get_payload(
                    payload_obj=self._tune_payload_obj,
                    input_classes=self._tune_obj.input_classes,
                    ignore_args=fixed_args,
                )
                # Build the Spockspace from the payload and the classes
                # Tuneable parameters
                self._tune_namespace = self._tune_obj.generate(self._tune_args)
        except Exception as e:
            self._print_usage_and_exit(str(e), sys_exit=False)
            raise ValueError(e)

    def __call__(self, *args, **kwargs):
        """Call to self to allow chaining

        *Args*:

            *args: non-keyword args
            **kwargs: keyword args

        *Returns*:

            ConfigArgBuilder: self instance
        """
        return ConfigArgBuilder(*args, **kwargs)

    def generate(self):
        """Generate method that returns the actual argument namespace


        *Returns*:

            argument namespace consisting of all config classes

        """
        return self._arg_namespace

    @property
    def tuner_status(self):
        """Returns a dictionary of all the necessary underlying tuner internals to report the result"""
        return self._tuner_status

    @property
    def best(self):
        """Returns a Spockspace of the best hyper-parameter config and the associated metric value"""
        return self._tuner_interface.best

    def sample(self):
        """Sample method that constructs a namespace from the fixed parameters and samples from the tuner space to
        generate a Spockspace derived from both

        *Returns*:

            argument namespace(s) -- fixed + drawn sample from tuner backend

        """
        if self._tune_obj is None:
            raise ValueError(
                f"Called sample method without passing any @spockTuner decorated classes"
            )
        if self._tuner_interface is None:
            raise ValueError(
                f"Called sample method without first calling the tuner method that initializes the "
                f"backend library"
            )
        return_tuple = self._tuner_state
        self._tuner_status = self._tuner_interface.tuner_status
        self._tuner_state = self._tuner_interface.sample()
        self._sample_count += 1
        return return_tuple

    def tuner(self, tuner_config):
        """Chained call that builds the tuner interface for either optuna or ax depending upon the type of the tuner_obj

        *Args*:

            tuner_config: a class of type optuna.study.Study or AX****

        *Returns*:

            self so that functions can be chained

        """
        if self._tune_obj is None:
            raise ValueError(
                f"Called tuner method without passing any @spockTuner decorated classes"
            )
        try:
            from spock.addons.tune.tuner import TunerInterface

            self._tuner_interface = TunerInterface(
                tuner_config=tuner_config,
                tuner_namespace=self._tune_namespace,
                fixed_namespace=self._arg_namespace,
            )
            self._tuner_state = self._tuner_interface.sample()
        except ImportError:
            print(
                "Missing libraries to support tune functionality. Please re-install with the extra tune "
                "dependencies -- pip install spock-config[tune]"
            )
        return self

    def _print_usage_and_exit(self, msg=None, sys_exit=True, exit_code=1):
        """Prints the help message and exits

        *Args*:

            msg: message to print pre exit

        *Returns*:

            None

        """
        print(f"usage: {sys.argv[0]} -c [--config] config1 [config2, config3, ...]")
        print(f'\n{self._desc if self._desc != "" else ""}\n')
        print("configuration(s):\n")
        # Call the fixed parameter help info
        self._builder_obj.handle_help_info()
        if self._tune_obj is not None:
            self._tune_obj.handle_help_info()
        if msg is not None:
            print(msg)
        if sys_exit:
            sys.exit(exit_code)

    @staticmethod
    def _handle_tuner_objects(tune_args, s3_config, kwargs):
        """Handles creating the tuner builder object if @spockTuner classes were passed in

        *Args*:

            tune_args: list of tuner classes
            s3_config: s3Config object for S3 support
            kwargs: optional keyword args

        *Returns*:

            tuner builder object or None

        """
        if len(tune_args) > 0:
            try:
                from spock.addons.tune.builder import TunerBuilder
                from spock.addons.tune.payload import TunerPayload

                tuner_builder = TunerBuilder(*tune_args, **kwargs)
                tuner_payload = TunerPayload(s3_config=s3_config)
                return tuner_builder, tuner_payload
            except ImportError:
                print(
                    "Missing libraries to support tune functionality. Please re-install with the extra tune "
                    "dependencies -- pip install spock-config[tune]"
                )
        else:
            return None, None

    @staticmethod
    def _verify_attr(args: typing.Tuple):
        """Verifies that all the input classes are attr based

        *Args*:

            args: tuple of classes passed to the builder

        *Returns*:

            None

        """
        # Gather if all attr backend
        type_attrs = all([attr.has(arg) for arg in args])
        if not type_attrs:
            which_idx = [attr.has(arg) for arg in args].index(False)
            if hasattr(args[which_idx], "__name__"):
                raise TypeError(
                    f"*args must be of all attrs backend -- missing a @spock decorator on class "
                    f"{args[which_idx].__name__}"
                )
            else:
                raise TypeError(
                    f"*args must be of all attrs backend -- invalid type "
                    f"{type(args[which_idx])}"
                )

    @staticmethod
    def _strip_tune_parameters(args: typing.Tuple):
        """Separates the fixed arguments from any hyper-parameter arguments

        *Args*:

            args: tuple of classes passed to the builder

        *Returns*:

            fixed_args: list of fixed args
            tune_args: list of args destined for a tuner backend

        """
        fixed_args = []
        tune_args = []
        for arg in args:
            if arg.__module__ == "spock.backend.config":
                fixed_args.append(arg)
            elif arg.__module__ == "spock.addons.tune.config":
                tune_args.append(arg)
        return fixed_args, tune_args

    def _handle_cmd_line(self):
        """Handle all cmd line related tasks

        Config paths can enter from either the command line or be added in the class init call
        as a kwarg (configs=[]) -- also trigger the building of the cmd line overrides for each fixed and
        tunable objects

        *Returns*:

            args: namespace of args

        """
        # Need to hold an overarching parser here that just gets appended to for both fixed and tunable objects
        # Check if the no_cmd_line is not flagged and if the configs are not empty
        if self._no_cmd_line and (self._configs is None):
            raise ValueError(
                "Flag set for preventing command line read but no paths were passed to the config kwarg"
            )
        # If cmd_line is flagged then build the parsers if not make any empty Namespace
        args = (
            self._build_override_parsers(desc=self._desc)
            if not self._no_cmd_line
            else argparse.Namespace(config=[], help=False)
        )
        # If configs are present from the init call then roll these into the namespace
        if self._configs is not None:
            args = self._get_from_kwargs(args, self._configs)
        return args

    def _build_override_parsers(self, desc):
        """Creates parsers for command-line overrides

        Builds the basic command line parser for configs and help then iterates through each attr instance to make
        namespace specific cmd line override parsers -- handles calling both the fixed and tunable objects

        *Args*:

            desc: argparser description

        *Returns*:

            args: argument namespace

        """
        # Highest level parser object
        parser = argparse.ArgumentParser(description=desc, add_help=False)
        parser.add_argument("-c", "--config", required=False, nargs="+", default=[])
        parser.add_argument("-h", "--help", action="store_true")
        # Handle the builder obj
        parser = self._builder_obj.build_override_parsers(parser=parser)
        if self._tune_obj is not None:
            parser = self._tune_obj.build_override_parsers(parser=parser)
        args = parser.parse_args()
        return args

    @staticmethod
    def _get_from_kwargs(args, configs):
        """Get configs from the configs kwarg

        *Args*:

            args: argument namespace
            configs: config kwarg

        *Returns*:

            args: arg namespace

        """
        if isinstance(configs, list):
            args.config.extend(configs)
        else:
            raise TypeError(
                f"configs kwarg must be of type list -- given {type(configs)}"
            )
        return args

    def _get_payload(self, payload_obj, input_classes, ignore_args: typing.List):
        """Get the parameter payload from the config file(s)

        Calls the various ways to get configs and then parses to retrieve the parameter payload - make sure to call
        deep update so as to not lose some parameters when only partially updating the payload

        *Args*:

            payload_obj: current payload object to call
            input_classes: classes to use to get payload
            ignore_args: args that were decorated for hyper-parameter tuning

        *Returns*:

            payload: dictionary of parameter values

        """
        if self._args.help:
            # Call sys exit with a clean code as this is the help call which is not unexpected behavior
            self._print_usage_and_exit(sys_exit=True, exit_code=0)
        payload = {}
        dependencies = {"paths": [], "rel_paths": [], "roots": []}
        if payload_obj is not None:
            # Make sure we are actually trying to map to input classes
            if len(input_classes) > 0:
                # If configs are present then iterate through them and deal with the payload
                if len(self._args.config) > 0:
                    for configs in self._args.config:
                        payload_update = payload_obj.payload(
                            input_classes,
                            ignore_args,
                            configs,
                            self._args,
                            dependencies,
                        )
                        check_payload_overwrite(payload, payload_update, configs)
                        deep_payload_update(payload, payload_update)
                # If there are no configs present we have to fall back only on cmd line args to fill out the necessary
                # data -- this is essentially using spock as a drop in replacement of arg-parser
                else:
                    payload_update = payload_obj.payload(
                        input_classes, ignore_args, None, self._args, dependencies
                    )
                    check_payload_overwrite(payload, payload_update, None)
                    deep_payload_update(payload, payload_update)
        return payload

    def _save(
        self,
        payload,
        file_name: str = None,
        user_specified_path: str = None,
        create_save_path: bool = True,
        extra_info: bool = True,
        file_extension: str = ".yaml",
        tuner_payload=None,
        fixed_uuid=None,
    ):
        """Private interface -- saves the current config setup to file with a UUID

        *Args*:

            payload: Spockspace to save
            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to just uuid if None
            user_specified_path: if user provides a path it will be used as the path to write
            create_save_path: bool to create the path to save if called
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)
            tuner_payload: tuner level payload (unsampled)
            fixed_uuid: fixed uuid to allow for file overwrite

        *Returns*:

            self so that functions can be chained
        """
        if user_specified_path is not None:
            save_path = Path(user_specified_path)
        elif self._builder_obj.save_path is not None:
            save_path = Path(self._builder_obj.save_path)
        else:
            raise ValueError(
                "Save did not receive a valid path from: (1) markup file(s) or (2) "
                "the keyword arg user_specified_path"
            )
        # Call the saver class and save function
        self._saver_obj.save(
            payload,
            save_path,
            file_name,
            create_save_path,
            extra_info,
            file_extension,
            tuner_payload,
            fixed_uuid,
        )
        return self

    def save(
        self,
        file_name: str = None,
        user_specified_path: str = None,
        create_save_path: bool = True,
        extra_info: bool = True,
        file_extension: str = ".yaml",
        add_tuner_sample: bool = False,
    ):
        """Saves the current config setup to file with a UUID

        *Args*:

            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to just uuid if None
            user_specified_path: if user provides a path it will be used as the path to write
            create_save_path: bool to create the path to save if called
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)
            append_tuner_state: save the current tuner sample to the payload

        *Returns*:

            self so that functions can be chained
        """
        if add_tuner_sample:
            if self._tune_obj is None:
                raise ValueError(
                    f"Called save method with add_tuner_sample as {add_tuner_sample} without passing any @spockTuner "
                    f"decorated classes -- please use the add_tuner_sample flag for saving only hyper-parameter tuning "
                    f"runs"
                )
            file_name = (
                f"hp.sample.{self._sample_count+1}"
                if file_name is None
                else f"{file_name}.hp.sample.{self._sample_count+1}"
            )
            self._save(
                self._tuner_state,
                file_name,
                user_specified_path,
                create_save_path,
                extra_info,
                file_extension,
            )
        else:
            self._save(
                self._arg_namespace,
                file_name,
                user_specified_path,
                create_save_path,
                extra_info,
                file_extension,
                tuner_payload=self._tune_namespace
                if self._tune_obj is not None
                else None,
            )
        return self

    def save_best(
        self,
        file_name: str = None,
        user_specified_path: str = None,
        create_save_path: bool = True,
        extra_info: bool = True,
        file_extension: str = ".yaml",
    ):
        """Saves the current best config setup to file

        *Args*:

            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to just uuid if None
            user_specified_path: if user provides a path it will be used as the path to write
            create_save_path: bool to create the path to save if called
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)

        *Returns*:

            self so that functions can be chained
        """
        if self._tune_obj is None:
            raise ValueError(
                f"Called save_best method without passing any @spockTuner decorated classes -- please use the save()"
                f" method for saving non hyper-parameter tuning runs"
            )
        file_name = f"hp.best" if file_name is None else f"{file_name}.hp.best"
        self._save(
            Spockspace(**vars(self._arg_namespace), **vars(self.best[0])),
            file_name,
            user_specified_path,
            create_save_path,
            extra_info,
            file_extension,
            fixed_uuid=self._fixed_uuid,
        )

        return self

    @property
    def config_2_dict(self):
        """Dictionary representation of the arg payload"""
        return self._saver_obj.dict_payload(self._arg_namespace)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles import aliases to allow backwards compat with backends"""

# from spock.backend.dataclass.args import *
from spock.backend.typed import SavePath



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles payloads from markup files"""

import os
import sys
from abc import abstractmethod
from itertools import chain
from pathlib import Path

from spock.backend.handler import BaseHandler
from spock.backend.utils import (
    convert_to_tuples,
    deep_update,
    get_attr_fields,
    get_type_fields,
)
from spock.utils import check_path_s3


class BasePayload(BaseHandler):  # pylint: disable=too-few-public-methods
    """Handles building the payload for config file(s)

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via recursive calls

    *Attributes*:

        _loaders: maps of each file extension to the loader class
        __s3_config: optional S3Config object to handle s3 access

    """

    def __init__(self, s3_config=None):
        super(BasePayload, self).__init__(s3_config=s3_config)

    @staticmethod
    @abstractmethod
    def _update_payload(base_payload, input_classes, ignore_classes, payload):
        """Updates the payload

        Checks the parameters defined in the config files against the provided classes and if
        passable adds them to the payload

        *Args*:

            base_payload: current payload
            input_classes: class to roll into
            ignore_classes: list of classes to ignore
            payload: total payload

        *Returns*:

            payload: updated payload

        """

    def payload(self, input_classes, ignore_classes, path, cmd_args, deps):
        """Builds the payload from config files

        Public exposed call to build the payload and set any command line overrides

        *Args*:

            input_classes: list of backend classes
            ignore_classes: list of classes to ignore
            path: path to config file(s)
            cmd_args: command line overrides
            deps: dictionary of config dependencies

        *Returns*:

            payload: dictionary of all mapped parameters

        """
        payload = self._payload(input_classes, ignore_classes, path, deps, root=True)
        payload = self._handle_overrides(payload, ignore_classes, cmd_args)
        return payload

    def _payload(self, input_classes, ignore_classes, path, deps, root=False):
        """Private call to construct the payload

        Main function call that builds out the payload from config files of multiple types. It handles
        various file types and also composition of config files via a recursive calls

        *Args*:
            input_classes: list of backend classes
            ignore_classes: list of classes to ignore
            path: path to config file(s)
            deps: dictionary of config dependencies

        *Returns*:

            payload: dictionary of all mapped parameters

        """
        # empty payload
        payload = {}
        if path is not None:
            # Match to loader based on file-extension
            config_extension = Path(path).suffix.lower()
            # Verify extension
            self._check_extension(file_extension=config_extension)
            # Load from file
            base_payload = self._supported_extensions.get(config_extension)().load(
                path, s3_config=self._s3_config
            )
            # Check and? update the dependencies
            deps = self._handle_dependencies(deps, path, root)
            if "config" in base_payload:
                payload = self._handle_includes(
                    base_payload,
                    config_extension,
                    input_classes,
                    ignore_classes,
                    path,
                    payload,
                    deps,
                )
            payload = self._update_payload(
                base_payload, input_classes, ignore_classes, payload
            )
        return payload

    @staticmethod
    def _handle_dependencies(deps, path, root):
        """Handles config file dependencies

        Checks to see if the config path (full or relative) has already been encountered. Essentially a DFS for graph
        cycles

        *Args*:

            deps: dictionary of config dependencies
            path: current config path
            root: boolean if root

        *Returns*:

            deps: updated dependencies

        """
        if root and path in deps.get("paths"):
            raise ValueError(
                f"Duplicate Read -- Config file {path} has already been encountered. "
                f"Please remove duplicate reads of config files."
            )
        elif path in deps.get("paths") or path in deps.get("rel_paths"):
            raise ValueError(
                f"Cyclical Dependency -- Config file {path} has already been encountered. "
                f"Please remove cyclical dependencies between config files."
            )
        else:
            # Update the dependency lists
            deps.get("paths").append(path)
            deps.get("rel_paths").append(os.path.basename(path))
            if root:
                deps.get("roots").append(path)
        return deps

    def _handle_includes(
        self,
        base_payload,
        config_extension,
        input_classes,
        ignore_classes,
        path,
        payload,
        deps,
    ):  # pylint: disable=too-many-arguments
        """Handles config composition

        For all of the config tags in the config file this function will recursively call the payload function
        with the composition path to get the additional payload(s) from the composed file(s) -- checks for file
        validity or if it is an S3 URI via regex

        *Args*:

            base_payload: base payload that has a config kwarg
            config_extension: file type
            input_classes: defined backend classes
            ignore_classes: list of classes to ignore
            path: path to base file
            payload: payload pulled from composed files
            deps: dictionary of config dependencies

        *Returns*:

            payload: payload update from composed files

        """
        included_params = {}
        for inc_path in base_payload["config"]:
            if check_path_s3(inc_path):
                use_path = inc_path
            elif os.path.exists(inc_path):
                use_path = inc_path
            elif os.path.join(os.path.dirname(path), inc_path):
                use_path = os.path.join(os.path.dirname(path), inc_path)
            else:
                raise RuntimeError(
                    f"Could not find included {config_extension} file {inc_path} or is not an S3 URI!"
                )
            included_params.update(
                self._payload(input_classes, ignore_classes, use_path, deps)
            )
        payload.update(included_params)
        return payload

    def _handle_overrides(self, payload, ignore_classes, args):
        """Handle command line overrides

        Iterate through the command line override values, determine at what level to set them, and set them if possible

        *Args*:

            payload: current payload dictionary
            args: command line override args

        *Returns*:

            payload: updated payload dictionary with override values set

        """
        skip_keys = ["config", "help"]
        pruned_args = self._prune_args(args, ignore_classes)
        for k, v in pruned_args.items():
            if k not in skip_keys and v is not None:
                payload = self._handle_payload_override(payload, k, v)
        return payload

    @staticmethod
    def _prune_args(args, ignore_classes):
        """Prunes ignored class names from the cmd line args list to prevent incorrect access

        *Args*:

            args: current cmd line args
            ignore_classes: list of class names to ignore

        *Returns*:

            dictionary of pruned cmd line args

        """
        ignored_stems = [val.__name__ for val in ignore_classes]
        return {
            k: v for k, v in vars(args).items() if k.split(".")[0] not in ignored_stems
        }

    @staticmethod
    @abstractmethod
    def _handle_payload_override(payload, key, value):
        """Handles the complex logic needed for List[spock class] overrides

        Messy logic that sets overrides for the various different types. The hardest being List[spock class] since str
        names have to be mapped backed to sys.modules and can be set at either the general or class level.

        *Args*:

            payload: current payload dictionary
            key: current arg key
            value: value at current arg key

        *Returns*:

            payload: modified payload with overrides

        """


class AttrPayload(BasePayload):
    """Handles building the payload for attrs backend

    This class builds out the payload from config files of multiple types. It handles various
    file types and also composition of config files via a recursive calls

    *Attributes*:

        _loaders: maps of each file extension to the loader class

    """

    def __init__(self, s3_config=None):
        """Init for AttrPayload

        *Args*:

            s3_config: optional S3 config object

        """
        super().__init__(s3_config=s3_config)

    def __call__(self, *args, **kwargs):
        """Call to allow self chaining

        *Args*:

            *args:
            **kwargs:

        *Returns*:

            Payload: instance of self

        """
        return AttrPayload(*args, **kwargs)

    @staticmethod
    def _update_payload(base_payload, input_classes, ignore_classes, payload):
        # Get basic args
        attr_fields = get_attr_fields(input_classes=input_classes)
        # Get the ignore fields
        ignore_fields = get_attr_fields(input_classes=ignore_classes)
        # Class names
        class_names = [val.__name__ for val in input_classes]
        # Parse out the types if generic
        type_fields = get_type_fields(input_classes)
        for keys, values in base_payload.items():
            if keys not in ignore_fields:
                # check if the keys, value pair is expected by the attr class
                if keys != "config":
                    # Dict infers that we are overriding a global setting in a specific config
                    if isinstance(values, dict):
                        # we're in a namespace
                        # Check for incorrect specific override of global def
                        if keys not in attr_fields:
                            raise TypeError(
                                f"Referring to a class space {keys} that is undefined"
                            )
                        for i_keys in values.keys():
                            if i_keys not in attr_fields[keys]:
                                raise ValueError(
                                    f"Provided an unknown argument named {keys}.{i_keys}"
                                )
                    else:
                        # Check if the key is actually a reference to another class
                        if keys in class_names:
                            if isinstance(values, list):
                                # Check for incorrect specific override of global def
                                if keys not in attr_fields:
                                    raise ValueError(
                                        f"Referring to a class space {keys} that is undefined"
                                    )
                                # We are in a repeated class def
                                # Raise if the key set is different from the defined set (i.e. incorrect arguments)
                                key_set = set(
                                    list(chain(*[list(val.keys()) for val in values]))
                                )
                                for i_keys in key_set:
                                    if i_keys not in attr_fields[keys]:
                                        raise ValueError(
                                            f"Provided an unknown argument named {keys}.{i_keys}"
                                        )
                            # Chain all the values from multiple spock classes into one list
                            elif keys not in list(chain(*attr_fields.values())):
                                raise ValueError(
                                    f"Provided an unknown argument named {keys}"
                                )
                        # Chain all the values from multiple spock classes into one list
                        elif keys not in list(chain(*attr_fields.values())):
                            raise ValueError(
                                f"Provided an unknown argument named {keys}"
                            )
                if keys in payload and isinstance(values, dict):
                    payload[keys].update(values)
                else:
                    payload[keys] = values
        tuple_payload = convert_to_tuples(payload, type_fields, class_names)
        payload = deep_update(payload, tuple_payload)
        return payload

    @staticmethod
    def _handle_payload_override(payload, key, value):
        """Handles the complex logic needed for List[spock class] overrides

        Messy logic that sets overrides for the various different types. The hardest being List[spock class] since str
        names have to be mapped backed to sys.modules and can be set at either the general or class level.

        *Args*:

            payload: current payload dictionary
            key: current arg key
            value: value at current arg key

        *Returns*:

            payload: modified payload with overrides

        """
        key_split = key.split(".")
        curr_ref = payload
        # Handle non existing parts of the payload for specific cases
        root_classes = [
            idx
            for idx, val in enumerate(key_split)
            if hasattr(sys.modules["spock"].backend.config, val)
        ]
        # Verify any classes have roots in the payload dict
        for idx in root_classes:
            # Update all root classes if not present
            if key_split[idx] not in payload:
                payload.update({key_split[idx]: {}})
            # If not updating the root then it is a reference to another class which might not be in the payload
            # Make sure it's there by setting it -- since this is an override setting is fine as these should be the
            # final say in the param values so don't worry about clashing
            if idx != 0:
                payload[key_split[0]][key_split[idx - 1]] = key_split[idx]
                # Check also for repeated classes -- value will be a list when the type is not
                var = getattr(
                    getattr(
                        sys.modules["spock"].backend.config, key_split[idx]
                    ).__attrs_attrs__,
                    key_split[-1],
                )
                if isinstance(value, list) and var.type != list:
                    # If the dict is blank we need to handle the creation of the list of dicts
                    if len(payload[key_split[idx]]) == 0:
                        payload.update(
                            {
                                key_split[idx]: [
                                    {key_split[-1]: None} for _ in range(len(value))
                                ]
                            }
                        )
                    # If it's already partially filled we need to update not overwrite
                    else:
                        for val in payload[key_split[idx]]:
                            val.update({key_split[-1]: None})

        for idx, split in enumerate(key_split):
            # Check for curr_ref switch over -- verify by checking the sys modules names
            if (
                idx != 0
                and (split in payload)
                and (isinstance(curr_ref, str))
                and (hasattr(sys.modules["spock"].backend.config, split))
            ):
                curr_ref = payload[split]
                # Look ahead to check if the next value exists in the dictionary
            elif (
                idx != 0
                and (split in payload)
                and (isinstance(payload[split], str))
                and (hasattr(sys.modules["spock"].backend.config, payload[split]))
            ):
                curr_ref = payload[split]
            # elif check if it's the last value and figure out the override
            elif idx == (len(key_split) - 1):
                # Handle bool(s) a bit differently as they are store_true
                if isinstance(curr_ref, dict) and isinstance(value, bool):
                    if value is not False:
                        curr_ref[split] = value
                # If we are at the dictionary level we should be able to just payload override
                elif isinstance(curr_ref, dict) and not isinstance(value, bool):
                    curr_ref[split] = value
                # If we are at a list level it must be some form of repeated class since this is the end of the class
                # tree -- check the instance type but also make sure the cmd-line override is the correct len
                elif isinstance(curr_ref, list) and len(value) == len(curr_ref):
                    # Walk the list and check for the key
                    for ref_idx, val in enumerate(curr_ref):
                        if split in val:
                            val[split] = value[ref_idx]
                        else:
                            raise ValueError(
                                f"cmd-line override failed for {key} -- "
                                f"Failed to find key {split} within lowest level List[Dict]"
                            )
                elif isinstance(curr_ref, list) and len(value) != len(curr_ref):
                    raise ValueError(
                        f"cmd-line override failed for {key} -- "
                        f"Specified key {split} with len {len(value)} does not match len {len(curr_ref)} "
                        f"of List[Dict]"
                    )
                else:
                    raise ValueError(
                        f"cmd-line override failed for {key} -- "
                        f"Failed to find key {split} within lowest level Dict"
                    )
            # If it's not keep walking the current payload
            else:
                curr_ref = curr_ref[split]
        return payload



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""I/O handlers for various file formats"""

import json
import os
import re
import typing
from abc import ABC, abstractmethod
from warnings import warn

import pytomlpp
import yaml

from spock import __version__
from spock.utils import check_path_s3


class Handler(ABC):
    """Base class for file type loaders

    ABC for loaders

    """

    def load(self, path: str, s3_config=None) -> typing.Dict:
        """Load function for file type

        This handles s3 path conversion for all handler types pre load call

        *Args*:

            path: path to file
            s3_config: optional s3 config object if using s3 storage

        *Returns*:

            dictionary of read file

        """
        path = self._handle_possible_s3_load_path(path=path, s3_config=s3_config)
        return self._load(path=path)

    @abstractmethod
    def _load(self, path: str) -> typing.Dict:
        """Private load function for file type

        *Args*:

            path: path to file

        *Returns*:

            dictionary of read file

        """
        raise NotImplementedError

    def save(
        self,
        out_dict: typing.Dict,
        info_dict: typing.Optional[typing.Dict],
        path: str,
        name: str,
        create_path: bool = False,
        s3_config=None,
    ):
        """Write function for file type

        This will handle local or s3 writes with the boolean is_s3 flag. If detected it will conditionally import
        the necessary addons to handle the upload

        *Args*:

            out_dict: payload to write
            info_dict: info payload to write
            path: path to write out
            name: spock generated file name
            create_path: boolean to create the path if non-existent (for non S3)
            s3_config: optional s3 config object if using s3 storage

        *Returns*:

        """
        write_path, is_s3 = self._handle_possible_s3_save_path(
            path=path, name=name, create_path=create_path, s3_config=s3_config
        )
        write_path = self._save(out_dict=out_dict, info_dict=info_dict, path=write_path)
        # After write check if it needs to be pushed to S3
        if is_s3:
            try:
                from spock.addons.s3.utils import handle_s3_save_path

                handle_s3_save_path(
                    temp_path=write_path, s3_path=path, name=name, s3_config=s3_config
                )
            except ImportError:
                print("Error importing spock s3 utils after detecting s3:// save path")

    @abstractmethod
    def _save(
        self, out_dict: typing.Dict, info_dict: typing.Optional[typing.Dict], path: str
    ) -> str:
        """Write function for file type

        *Args*:

            out_dict: payload to write
            info_dict: info payload to write
            path: path to write out

        *Returns*:

        """
        raise NotImplementedError

    @staticmethod
    def _handle_possible_s3_load_path(path: str, s3_config=None) -> str:
        """Handles the possibility of having to handle loading from a S3 path

        Checks to see if it detects a S3 uri and if so triggers imports of s3 functionality and handles the file
        download

        *Args*:

            path: spock config path
            s3_config: optional s3 configuration object

        *Returns*:

            path: current path for the configuration file

        """
        is_s3 = check_path_s3(path=path)
        if is_s3:
            try:
                from spock.addons.s3.utils import handle_s3_load_path

                path = handle_s3_load_path(path=path, s3_config=s3_config)
            except ImportError:
                print("Error importing spock s3 utils after detecting s3:// load path")
        return path

    @staticmethod
    def _handle_possible_s3_save_path(
        path: str, name: str, create_path: bool, s3_config=None
    ) -> typing.Tuple[str, bool]:
        """Handles the possibility of having to save to a S3 path

        Checks to see if it detects a S3 uri and if so generates a tmp location to write the file to pre-upload

        *Args*:

            path: save path
            name: spock generated file name
            create_path: create the path for non s3 data
            s3_config: s3 config object

        *Returns*:

        """
        is_s3 = check_path_s3(path=path)
        if is_s3:
            if s3_config is None:
                raise ValueError(
                    "Save to S3 -- Missing S3Config object which is necessary to handle S3 style paths"
                )
            write_path = f"{s3_config.temp_folder}/{name}"
            # Strip double slashes if exist
            write_path = write_path.replace(r"//", r"/")
        else:
            # Handle the path logic for non S3
            if not os.path.exists(path) and create_path:
                os.makedirs(path)
            write_path = f"{path}/{name}"
        return write_path, is_s3

    @staticmethod
    def write_extra_info(path, info_dict):
        """Writes extra info to commented newlines

        *Args*:

            path: path to write out
            info_dict: info payload to write

        *Returns*:

        """
        # Write the commented info as new lines
        with open(path, "w+") as fid:
            # Write a spock header
            fid.write(f"# Spock Version: {__version__}\n")
            # Write info dict if not None
            if info_dict is not None:
                for k, v in info_dict.items():
                    fid.write(f"{k}: {v}\n")
            fid.write("\n")


class YAMLHandler(Handler):
    """YAML class for loading YAML config files

    Base YAML class

    """

    # override default SafeLoader behavior to correctly
    # interpret 1e1 (as opposed to 1.e+1) as 10
    # https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number/30462009#30462009
    yaml.SafeLoader.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
         [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )

    def _load(self, path: str) -> typing.Dict:
        """YAML load function

        *Args*:

            path: path to YAML file

        *Returns*:

            base_payload: dictionary of read file

        """
        file_contents = open(path, "r").read()
        file_contents = re.sub(r"--([a-zA-Z0-9_]*)", r"\g<1>: True", file_contents)
        base_payload = yaml.safe_load(file_contents)
        return base_payload

    def _save(
        self, out_dict: typing.Dict, info_dict: typing.Optional[typing.Dict], path: str
    ):
        """Write function for YAML type

        *Args*:

            out_dict: payload to write
            info_dict: info payload to write
            path: path to write out

        *Returns*:

        """
        # First write the commented info
        self.write_extra_info(path=path, info_dict=info_dict)
        # Remove aliases in YAML dump
        yaml.Dumper.ignore_aliases = lambda *args: True
        with open(path, "a") as yaml_fid:
            yaml.safe_dump(out_dict, yaml_fid, default_flow_style=False)
        return path


class TOMLHandler(Handler):
    """TOML class for loading TOML config files

    Base TOML class

    """

    def _load(self, path: str) -> typing.Dict:
        """TOML load function

        *Args*:

            path: path to TOML file

        Returns:

            base_payload: dictionary of read file

        """
        base_payload = pytomlpp.load(path)
        return base_payload

    def _save(
        self, out_dict: typing.Dict, info_dict: typing.Optional[typing.Dict], path: str
    ):
        """Write function for TOML type

        *Args*:

            out_dict: payload to write
            info_dict: info payload to write
            path: path to write out

        *Returns*:

        """
        # First write the commented info
        self.write_extra_info(path=path, info_dict=info_dict)
        with open(path, "a") as toml_fid:
            pytomlpp.dump(out_dict, toml_fid)
        return path


class JSONHandler(Handler):
    """JSON class for loading JSON config files

    Base JSON class

    """

    def _load(self, path: str) -> typing.Dict:
        """JSON load function

        *Args*:

            path: path to JSON file

        Returns:

            base_payload: dictionary of read file

        """
        with open(path) as json_fid:
            base_payload = json.load(json_fid)
        return base_payload

    def _save(
        self, out_dict: typing.Dict, info_dict: typing.Optional[typing.Dict], path: str
    ):
        """Write function for JSON type

        *Args*:

            out_dict: payload to write
            info_dict: info payload to write
            path: path to write out

        *Returns*:

        """
        if info_dict is not None:
            warn(
                "JSON does not support comments and thus cannot save extra info to file... removing extra info"
            )
            info_dict = None
        with open(path, "a") as json_fid:
            json.dump(out_dict, json_fid, indent=4, separators=(",", ": "))
        return path



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Attr utility functions for Spock"""


def get_attr_fields(input_classes):
    """Gets the attribute fields from all classes

    *Args*:

        input_classes: current list of input classes

    *Returns*:

        dictionary of all attrs attribute fields

    """
    return {
        attr.__name__: [val.name for val in attr.__attrs_attrs__]
        for attr in input_classes
    }


def get_type_fields(input_classes):
    """Creates a dictionary of names and types

    *Args*:

        input_classes: list of input classes

    *Returns*:

        type_fields: dictionary of names and types

    """
    # Parse out the types if generic
    type_fields = {}
    for attr in input_classes:
        input_attr = {}
        for val in attr.__attrs_attrs__:
            if "type" in val.metadata:
                input_attr.update({val.name: val.metadata["type"]})
            else:
                input_attr.update({val.name: None})
        type_fields.update({attr.__name__: input_attr})
    return type_fields


def flatten_type_dict(type_dict):
    """Flattens a nested dictionary

    *Args*:

        type_dict: dictionary of types that are generic

    *Returns*:

        flat_dict: flatten dictionary to a single level

    """
    flat_dict = {}
    for k, v in type_dict.items():
        if isinstance(v, dict):
            return_dict = flatten_type_dict(v)
            flat_dict.update(return_dict)
        else:
            flat_dict[k] = v
    return flat_dict


def convert_to_tuples(input_dict, named_type_dict, class_names):
    """Convert lists to tuples

    Payloads from markup come in as Lists and not Tuples. This function turns lists in to tuples for the payloads
    so the attr values are set correctly. Will call itself recursively if a dictionary is present for class specific
    values

    *Args*:

        input_dict: input dictionary
        named_type_dict: dictionary of names with generic types

    *Returns*:

        updated_dict: a dictionary with lists converted to tuples

    """
    updated_dict = {}
    all_typed_dict = flatten_type_dict(named_type_dict)
    for k, v in input_dict.items():
        if k != "config":
            if isinstance(v, dict):
                updated = convert_to_tuples(v, named_type_dict.get(k), class_names)
                if updated:
                    updated_dict.update({k: updated})
            elif isinstance(v, list) and k in class_names:
                for val in v:
                    updated = convert_to_tuples(
                        val, named_type_dict.get(k), class_names
                    )
                    if updated:
                        updated_dict.update({k: updated})
            elif all_typed_dict[k] is not None:
                updated = _recursive_list_to_tuple(v, all_typed_dict[k], class_names)
                updated_dict.update({k: updated})
    return updated_dict


def deep_update(source, updates):
    """Deeply updates a dictionary

    Iterates through a dictionary recursively to update individual values within a possibly nested dictionary
    of dictionaries

    *Args*:

        source: source dictionary
        updates: updates to the dictionary

    *Returns*:

        source: updated version of the source dictionary

    """
    for k, v in updates.items():
        if isinstance(v, dict) and v:
            updated_dict = deep_update(source.get(k), v)
            if updated_dict:
                source[k] = updated_dict
        else:
            source[k] = v
    return source


def _recursive_list_to_tuple(value, typed, class_names):
    """Recursively turn lists into tuples

    Recursively looks through a pair of value and type and sets any of the possibly nested type of value to tuple
    if tuple is the specified type

    *Args*:

        value: value to check and set typ if necessary
        typed: type of the generic alias to check against

    *Returns*:

        value: updated value with correct type casts

    """
    # Check for __args__ as it signifies a generic and make sure it's not already been cast as a tuple
    # from a composed payload
    if (
        hasattr(typed, "__args__")
        and not isinstance(value, tuple)
        and not (isinstance(value, str) and value in class_names)
    ):
        # Force those with origin tuple types to be of the defined length
        if (typed.__origin__.__name__.lower() == "tuple") and len(value) != len(
            typed.__args__
        ):
            raise ValueError(
                f"Tuple(s) use a fixed/defined length -- Length of the provided argument ({len(value)}) "
                f"does not match the length of the defined argument ({len(typed.__args__)})"
            )
        # need to recurse before casting as we can't set values in a tuple with idx
        # Since it's generic it should be iterable to recurse and check it's children
        for idx, val in enumerate(value):
            value[idx] = _recursive_list_to_tuple(val, typed.__args__[0], class_names)
        # First check if list and then swap to tuple if the origin is tuple
        if isinstance(value, list) and typed.__origin__.__name__.lower() == "tuple":
            value = tuple(value)
    else:
        return value
    return value



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

import re
import sys
from abc import ABC, abstractmethod
from enum import EnumMeta
from typing import List

import attr
from attr import NOTHING

from spock.backend.wrappers import Spockspace
from spock.utils import make_argument


class BaseBuilder(ABC):  # pylint: disable=too-few-public-methods
    """Base class for building the backend specific builders

    This class handles the interface to the backend with the generic ConfigArgBuilder so that different
    backends can be used to handle processing

    *Attributes*

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        _max_indent: maximum to indent between help prints
        save_path: list of path(s) to save the configs to

    """

    def __init__(self, *args, max_indent=4, module_name, **kwargs):
        self.input_classes = args
        self._module_name = module_name
        self._max_indent = max_indent
        self.save_path = None

    @staticmethod
    @abstractmethod
    def _make_group_override_parser(parser, class_obj, class_name):
        """Makes a name specific override parser for a given class obj

        Takes a class object of the backend and adds a new argument group with argument names given with name
        Class.name so that individual parameters specific to a class can be overridden.

        *Args*:

            parser: argument parser
            class_obj: instance of a backend class
            class_name: used for module matching

        *Returns*:

            parser: argument parser with new class specific overrides

        """

    def handle_help_info(self):
        """Handles walking through classes to get help info

        For each class this function will search __doc__ and attempt to pull out help information for both the class
        itself and each attribute within the class

        *Returns*:

            None

        """
        self._attrs_help(self.input_classes, self._module_name)

    def _handle_arguments(self, args, class_obj):
        """Handles all argument mapping

        Creates a dictionary of named parameters that are mapped to the final type of object

        *Args*:

            args: read file arguments
            class_obj: instance of a class obj

        *Returns*:

            fields: dictionary of mapped parameters

        """
        attr_name = class_obj.__name__
        class_names = [val.__name__ for val in self.input_classes]
        # Handle repeated classes
        if (
            attr_name in class_names
            and attr_name in args
            and isinstance(args[attr_name], list)
        ):
            fields = self._handle_repeated(args[attr_name], attr_name, class_names)
        # Handle non-repeated classes
        else:
            fields = {}
            for val in class_obj.__attrs_attrs__:
                # Check if namespace is named and then check for key -- checking for local class def
                if attr_name in args and val.name in args[attr_name]:
                    fields[val.name] = self._handle_nested_class(
                        args, args[attr_name][val.name], class_names
                    )
                # If not named then just check for keys -- checking for global def
                elif val.name in args:
                    fields[val.name] = self._handle_nested_class(
                        args, args[val.name], class_names
                    )
                # Check for special keys to set
                if (
                    "special_key" in val.metadata
                    and val.metadata["special_key"] is not None
                ):
                    if val.name in args:
                        self.save_path = args[val.name]
                    elif val.default is not None:
                        self.save_path = val.default
        return fields

    def _handle_repeated(self, args, check_value, class_names):
        """Handles repeated classes as lists

        *Args*:

            args: dictionary of arguments from the configs
            check_value: value to check classes against
            class_names: current class names

        *Returns*:

            list of input_class[match)idx[0]] types filled with repeated values

        """
        # Check to see if the value trying to be set is actually an input class
        match_idx = [idx for idx, val in enumerate(class_names) if val == check_value]
        return [self.input_classes[match_idx[0]](**val) for val in args]

    def _handle_nested_class(self, args, check_value, class_names):
        """Handles passing another class to the field dictionary

        *Args*:
            args: dictionary of arguments from the configs
            check_value: value to check classes against
            class_names: current class names

        *Returns*:

            either the check_value or the necessary class

        """
        # Check to see if the value trying to be set is actually an input class
        match_idx = [idx for idx, val in enumerate(class_names) if val == check_value]
        # If so then create the needed class object by unrolling the args to **kwargs and return it
        if len(match_idx) > 0:
            if len(match_idx) > 1:
                raise ValueError(
                    "Match error -- multiple classes with the same name definition"
                )
            else:
                if args.get(self.input_classes[match_idx[0]].__name__) is None:
                    raise ValueError(
                        f"Missing config file definition for the referenced class "
                        f"{self.input_classes[match_idx[0]].__name__}"
                    )
                current_arg = args.get(self.input_classes[match_idx[0]].__name__)
                if isinstance(current_arg, list):
                    class_value = [
                        self.input_classes[match_idx[0]](**val) for val in current_arg
                    ]
                else:
                    class_value = self.input_classes[match_idx[0]](**current_arg)
            return_value = class_value
        # else return the expected value
        else:
            return_value = check_value
        return return_value

    def generate(self, dict_args):
        """Method to auto-generate the actual class instances from the generated args

        Based on the generated arguments groups and the args read in from the config file(s)
        this function instantiates the classes with the necessary field or attr values

        *Args*:

            dict_args: dictionary of arguments from the configs

        *Returns*:

            namespace containing automatically generated instances of the classes
        """
        auto_dict = {}
        for attr_classes in self.input_classes:
            attr_build = self._auto_generate(dict_args, attr_classes)
            if isinstance(attr_build, list):
                class_name = list({type(val).__name__ for val in attr_build})
                if len(class_name) > 1:
                    raise ValueError("Repeated class has more than one unique name")
                auto_dict.update({class_name[0]: attr_build})
            else:
                auto_dict.update({type(attr_build).__name__: attr_build})
        return Spockspace(**auto_dict)

    def _auto_generate(self, args, input_class):
        """Builds an instance of an attr class

        Builds an instance with the necessary field values from the argument
        dictionary read from the config file(s)

        *Args*:

            args: dictionary of arguments read from the config file(s)
            data_class: data class to build

        *Returns*:

            An instance of data_class with correct values assigned to fields
        """
        # Handle the basic data types
        fields = self._handle_arguments(args, input_class)
        if isinstance(fields, list):
            return_value = fields
        else:
            self._handle_late_defaults(args, fields, input_class)
            return_value = input_class(**fields)
        return return_value

    def _handle_late_defaults(self, args, fields, input_class):
        """Handles late defaults when the type is non-standard

        If the default type is not a base python type then we need to catch those defaults here and build the correct
        values from the input classes while maintaining the optional nature. The trick is to exclude all 'base' types
        as these defaults are covered by the attr default value

        *Args*:

            args: dictionary of arguments read from the config file(s)
            fields: current fields returned from _handle_arguments
            input_class: which input class being checked for late defaults

        *Returns*:

            fields: updated field dictionary with late defaults set

        """
        names = [val.name for val in input_class.__attrs_attrs__]
        class_names = [val.__name__ for val in self.input_classes]
        field_list = list(fields.keys())
        arg_list = list(args.keys())
        # Exclude all the base types that are supported -- these can be set by attrs
        exclude_list = [
            "_Nothing",
            "NoneType",
            "bool",
            "int",
            "float",
            "str",
            "list",
            "tuple",
        ]
        for val in names:
            if val not in field_list:
                default_type_name = type(
                    getattr(input_class.__attrs_attrs__, val).default
                ).__name__
                if default_type_name not in exclude_list:
                    default_attr = getattr(input_class.__attrs_attrs__, val).default
                    # If the default is given for a class then it's the actual class and not a type -- logic needs
                    # to deal with both
                    if type(default_attr).__name__ == "type":
                        default_name = default_attr.__name__
                    else:
                        default_name = type(default_attr).__name__
                else:
                    default_name = None
                if default_name is not None and default_name in arg_list:
                    if isinstance(args.get(default_name), list):
                        default_value = [
                            self.input_classes[class_names.index(default_name)](
                                **arg_val
                            )
                            for arg_val in args.get(default_name)
                        ]
                    else:
                        default_value = self.input_classes[
                            class_names.index(default_name)
                        ](**args.get(default_name))
                    fields.update({val: default_value})
        return fields

    def build_override_parsers(self, parser):
        """Creates parsers for command-line overrides

        Builds the basic command line parser for configs and help then iterates through each attr instance to make
        namespace specific cmd line override parsers

        *Args*:

            parser: argument parser

        *Returns*:

            parser: argument parser with new class specific overrides

        """
        # Build out each class override specific parser
        for val in self.input_classes:
            parser = self._make_group_override_parser(
                parser=parser, class_obj=val, class_name=self._module_name
            )
        return parser

    @staticmethod
    def _get_from_kwargs(args, configs):
        """Get configs from the configs kwarg

        *Args*:

            args: argument namespace
            configs: config kwarg

        *Returns*:

            args: arg namespace

        """
        if isinstance(configs, list):
            args.config.extend(configs)
        else:
            raise TypeError(
                f"configs kwarg must be of type list -- given {type(configs)}"
            )
        return args

    @staticmethod
    def _find_attribute_idx(newline_split_docs):
        """Finds the possible split between the header and Attribute annotations

        *Args*:

            newline_split_docs: new line split text

        Returns:

            idx: -1 if none or the idx of Attributes

        """
        for idx, val in enumerate(newline_split_docs):
            re_check = re.search(r"(?i)Attribute?s?:", val)
            if re_check is not None:
                return idx
        return -1

    def _split_docs(self, obj):
        """Possibly splits head class doc string from attribute docstrings

        Attempts to find the first contiguous line within the Google style docstring to use as the class docstring.
        Splits the docs base on the Attributes tag if present.

        *Args*:

            obj: class object to rip info from

        *Returns*:

            class_doc: class docstring if present or blank str
            attr_doc: list of attribute doc strings

        """
        if obj.__doc__ is not None:
            # Split by new line
            newline_split_docs = obj.__doc__.split("\n")
            # Cleanup l/t whitespace
            newline_split_docs = [val.strip() for val in newline_split_docs]
        else:
            newline_split_docs = []
        # Find the break between the class docs and the Attribute section -- if this returns -1 then there is no
        # Attributes section
        attr_idx = self._find_attribute_idx(newline_split_docs)
        head_docs = (
            newline_split_docs[:attr_idx] if attr_idx != -1 else newline_split_docs
        )
        attr_docs = newline_split_docs[attr_idx:] if attr_idx != -1 else []
        # Grab only the first contiguous line as everything else will probably be too verbose (e.g. the
        # mid-level docstring that has detailed descriptions
        class_doc = ""
        for idx, val in enumerate(head_docs):
            class_doc += f" {val}"
            if idx + 1 != len(head_docs) and head_docs[idx + 1] == "":
                break
        # Clean up any l/t whitespace
        class_doc = class_doc.strip()
        if len(class_doc) > 0:
            class_doc = f"-- {class_doc}"
        return class_doc, attr_docs

    @staticmethod
    def _match_attribute_docs(
        attr_name, attr_docs, attr_type_str, attr_default=NOTHING
    ):
        """Matches class attributes with attribute docstrings via regex

        *Args*:

            attr_name: attribute name
            attr_docs: list of attribute docstrings
            attr_type_str: str representation of the attribute type
            attr_default: str representation of a possible default value

        *Returns*:

            dictionary of packed attribute information

        """
        # Regex match each value
        a_str = None
        for a_doc in attr_docs:
            match_re = re.search(r"(?i)^" + attr_name + "?:", a_doc)
            # Find only the first match -- if more than one than ignore
            if match_re:
                a_str = a_doc[match_re.end() :].strip()
        return {
            attr_name: {
                "type": attr_type_str,
                "desc": a_str if a_str is not None else "",
                "default": "(default: " + repr(attr_default) + ")"
                if type(attr_default).__name__ != "_Nothing"
                else "",
                "len": {"name": len(attr_name), "type": len(attr_type_str)},
            }
        }

    def _handle_attributes_print(self, info_dict):
        """Prints attribute information in an argparser style format

        *Args*:

            info_dict: packed attribute info dictionary to print

        """
        # Figure out indents
        max_param_length = max([len(k) for k in info_dict.keys()])
        max_type_length = max([v["len"]["type"] for v in info_dict.values()])
        # Print akin to the argparser
        for k, v in info_dict.items():
            print(
                f"    {k}"
                + (" " * (max_param_length - v["len"]["name"] + self._max_indent))
                + f'{v["type"]}'
                + (" " * (max_type_length - v["len"]["type"] + self._max_indent))
                + f'{v["desc"]} {v["default"]}'
            )
        # Blank for spacing :-/
        print("")

    def _extract_other_types(self, typed, module_name):
        """Takes a high level type and recursively extracts any enum or class types

        *Args*:

            typed: highest level type
            module_name: name of module to match

        *Returns*:

            return_list: list of nums (dot notation of module_path.enum_name or module_path.class_name)

        """
        return_list = []
        if hasattr(typed, "__args__"):
            for val in typed.__args__:
                recurse_return = self._extract_other_types(val, module_name)
                if isinstance(recurse_return, list):
                    return_list.extend(recurse_return)
                else:
                    return_list.append(self._extract_other_types(val, module_name))
        elif isinstance(typed, EnumMeta) or (typed.__module__ == module_name):
            return [f"{typed.__module__}.{typed.__name__}"]
        return return_list

    def _attrs_help(self, input_classes, module_name):
        """Handles walking through a list classes to get help info

        For each class this function will search __doc__ and attempt to pull out help information for both the class
        itself and each attribute within the class. If it finds a repeated class in a iterable object it will
        recursively call self to handle information

        *Args*:

            input_classes: list of attr classes
            module_name: name of module to match

        *Returns*:

            None

        """
        # Handle the main loop
        other_list = self._handle_help_main(input_classes, module_name)
        self._handle_help_enums(other_list=other_list, module_name=module_name)

    @staticmethod
    def _get_type_string(val, nested_others):
        """Gets the type of the attr val as a string

        *Args*:

            val: current attr being processed
            nested_others: list of nested others to deal with that might have module path info in the string

        *Returns*:

            type_string: type of the attr as a str

        """
        # Grab the base or type info depending on what is provided
        if "type" in val.metadata:
            type_string = repr(val.metadata["type"])
        elif "base" in val.metadata:
            type_string = val.metadata["base"]
        elif hasattr(val.type, "__name__"):
            type_string = val.type.__name__
        else:
            type_string = str(val.type)
        # Regex out the typing info if present
        type_string = re.sub(r"typing.", "", type_string)
        # Regex out any nested_others that have module path information
        for other_val in nested_others:
            split_other = f"{'.'.join(other_val.split('.')[:-1])}."
            type_string = re.sub(split_other, "", type_string)
        # Regex the string to see if it matches any Enums in the __main__ module space
        # Construct the type with the metadata
        if "optional" in val.metadata:
            type_string = f"Optional[{type_string}]"
        return type_string

    def _handle_help_main(self, input_classes, module_name):
        """Handles the print of the main class types

        *Args*:

            input_classes: current set of input classes
            module_name: module name to match

        *Returns*:

            other_list: extended list of other classes/enums to process

        """
        # List to catch Enums and classes and handle post spock wrapped attr classes
        other_list = []
        covered_set = set()
        for attrs_class in input_classes:
            # Split the docs into class docs and any attribute docs
            class_doc, attr_docs = self._split_docs(attrs_class)
            print("  " + attrs_class.__name__ + f" {class_doc}")
            # Keep a running info_dict of all the attribute level info
            info_dict = {}
            for val in attrs_class.__attrs_attrs__:
                # If the type is an enum we need to handle it outside of this attr loop
                # Match the style of nested enums and return a string of module.name notation
                if isinstance(val.type, EnumMeta):
                    other_list.append(f"{val.type.__module__}.{val.type.__name__}")
                # if there is a type (implied Iterable) -- check it for nested Enums or classes
                nested_others = self._extract_fnc(val, module_name)
                if len(nested_others) > 0:
                    other_list.extend(nested_others)
                # Get the type represented as a string
                type_string = self._get_type_string(val, nested_others)
                info_dict.update(
                    self._match_attribute_docs(
                        val.name, attr_docs, type_string, val.default
                    )
                )
            # Add to covered so we don't print help twice in the case of some recursive nesting
            covered_set.add(f"{attrs_class.__module__}.{attrs_class.__name__}")
            self._handle_attributes_print(info_dict=info_dict)
        # Convert the enum list to a set to remove dupes and then back to a list so it is iterable -- set diff to not
        # repeat
        return list(set(other_list) - covered_set)

    def _handle_help_enums(self, other_list, module_name):
        """handles any extra enums from non main args

        *Args*:

            other_list: extended list of other classes/enums to process
            module_name: module name to match

        *Returns*:

            None

        """
        # Iterate any Enum type classes
        for other in other_list:
            # if it's longer than 2 then it's an embedded Spock class
            if ".".join(other.split(".")[:-1]) == module_name:
                class_type = self._get_from_sys_modules(other)
                # Invoke recursive call for the class
                self._attrs_help([class_type], module_name)
            # Fall back to enum style
            else:
                enum = self._get_from_sys_modules(other)
                # Split the docs into class docs and any attribute docs
                class_doc, attr_docs = self._split_docs(enum)
                print("  " + enum.__name__ + f" ({class_doc})")
                info_dict = {}
                for val in enum:
                    info_dict.update(
                        self._match_attribute_docs(
                            attr_name=val.name,
                            attr_docs=attr_docs,
                            attr_type_str=type(val.value).__name__,
                        )
                    )
                self._handle_attributes_print(info_dict=info_dict)

    @abstractmethod
    def _extract_fnc(self, val, module_name):
        """Function that gets the nested lists within classes

        *Args*:

            val: current attr
            module_name: matching module name

        *Returns*:

            list of any nested classes/enums

        """

    @staticmethod
    def _get_from_sys_modules(cls_name):
        """Gets the class from a dot notation name

        *Args*:

            cls_name: dot notation enum name

        *Returns*:

            module: enum class

        """
        # Split on dot notation
        split_string = cls_name.split(".")
        module = None
        for idx, val in enumerate(split_string):
            # idx = 0 will always be a call to the sys.modules dict
            if idx == 0:
                module = sys.modules[val]
            # all other idx are paths along the module that need to be traversed
            # idx = -1 will always be the final Enum object name we want to grab (final getattr call)
            else:
                module = getattr(module, val)
        return module


class AttrBuilder(BaseBuilder):
    """Attr specific builder

    Class that handles building for the attr backend

    *Attributes*

        input_classes: list of input classes that link to a backend
        _configs: None or List of configs to read from
        _create_save_path: boolean to make the path to save to
        _desc: description for the arg parser
        _no_cmd_line: flag to force no command line reads
        save_path: list of path(s) to save the configs to

    """

    def __init__(self, *args, **kwargs):
        """AttrBuilder init

        Args:
            *args: list of input classes that link to a backend
            configs: None or List of configs to read from
            desc: description for the arg parser
            no_cmd_line: flag to force no command line reads
            **kwargs: any extra keyword args
        """
        super().__init__(*args, module_name="spock.backend.config", **kwargs)

    @staticmethod
    def _make_group_override_parser(parser, class_obj, class_name):
        """Makes a name specific override parser for a given class obj

        Takes a class object of the backend and adds a new argument group with argument names given with name
        Class.name so that individual parameters specific to a class can be overridden.

        *Args*:

            parser: argument parser
            class_obj: instance of a backend class
            class_name: used for module matching

        *Returns*:

            parser: argument parser with new class specific overrides

        """
        attr_name = class_obj.__name__
        group_parser = parser.add_argument_group(
            title=str(attr_name) + " Specific Overrides"
        )
        for val in class_obj.__attrs_attrs__:
            val_type = val.metadata["type"] if "type" in val.metadata else val.type
            # Check if the val type has __args__ -- this catches lists?
            # TODO (ncilfone): Fix up this super super ugly logic
            if (
                hasattr(val_type, "__args__")
                and ((list(set(val_type.__args__))[0]).__module__ == class_name)
                and attr.has((list(set(val_type.__args__))[0]))
            ):
                args = list(set(val_type.__args__))[0]
                for inner_val in args.__attrs_attrs__:
                    arg_name = f"--{str(attr_name)}.{val.name}.{args.__name__}.{inner_val.name}"
                    group_parser = make_argument(
                        arg_name, List[inner_val.type], group_parser
                    )
            # If it's a reference to a class it needs to be an arg of a simple string as class matching will take care
            # of it later on
            elif val_type.__module__ == "spock.backend.config":
                arg_name = f"--{str(attr_name)}.{val.name}"
                val_type = str
                group_parser = make_argument(arg_name, val_type, group_parser)
            else:
                arg_name = f"--{str(attr_name)}.{val.name}"
                group_parser = make_argument(arg_name, val_type, group_parser)
        return parser

    def _handle_arguments(self, args, class_obj):
        attr_name = class_obj.__name__
        class_names = [val.__name__ for val in self.input_classes]
        # Handle repeated classes
        if (
            attr_name in class_names
            and attr_name in args
            and isinstance(args[attr_name], list)
        ):
            fields = self._handle_repeated(args[attr_name], attr_name, class_names)
        # Handle non-repeated classes
        else:
            fields = {}
            for val in class_obj.__attrs_attrs__:
                # Check if namespace is named and then check for key -- checking for local class def
                if attr_name in args and val.name in args[attr_name]:
                    fields[val.name] = self._handle_nested_class(
                        args, args[attr_name][val.name], class_names
                    )
                # If not named then just check for keys -- checking for global def
                elif val.name in args:
                    fields[val.name] = self._handle_nested_class(
                        args, args[val.name], class_names
                    )
                # Check for special keys to set
                if (
                    "special_key" in val.metadata
                    and val.metadata["special_key"] is not None
                ):
                    if val.name in args:
                        self.save_path = args[val.name]
                    elif val.default is not None:
                        self.save_path = val.default
        return fields

    def _handle_repeated(self, args, check_value, class_names):
        """Handles repeated classes as lists

        *Args*:

            args: dictionary of arguments from the configs
            check_value: value to check classes against
            class_names: current class names

        *Returns*:

            list of input_class[match)idx[0]] types filled with repeated values

        """
        # Check to see if the value trying to be set is actually an input class
        match_idx = [idx for idx, val in enumerate(class_names) if val == check_value]
        return [self.input_classes[match_idx[0]](**val) for val in args]

    def _handle_nested_class(self, args, check_value, class_names):
        """Handles passing another class to the field dictionary

        *Args*:
            args: dictionary of arguments from the configs
            check_value: value to check classes against
            class_names: current class names

        *Returns*:

            either the check_value or the necessary class

        """
        # Check to see if the value trying to be set is actually an input class
        match_idx = [idx for idx, val in enumerate(class_names) if val == check_value]
        # If so then create the needed class object by unrolling the args to **kwargs and return it
        if len(match_idx) > 0:
            if len(match_idx) > 1:
                raise ValueError(
                    "Match error -- multiple classes with the same name definition"
                )
            else:
                if args.get(self.input_classes[match_idx[0]].__name__) is None:
                    raise ValueError(
                        f"Missing config file definition for the referenced class "
                        f"{self.input_classes[match_idx[0]].__name__}"
                    )
                current_arg = args.get(self.input_classes[match_idx[0]].__name__)
                if isinstance(current_arg, list):
                    class_value = [
                        self.input_classes[match_idx[0]](**val) for val in current_arg
                    ]
                else:
                    class_value = self.input_classes[match_idx[0]](**current_arg)
            return_value = class_value
        # else return the expected value
        else:
            return_value = check_value
        return return_value

    def _extract_fnc(self, val, module_name):
        """Function that gets the nested lists within classes

        *Args*:

            val: current attr
            module_name: matching module name

        *Returns*:

            list of any nested classes/enums

        """
        return (
            self._extract_other_types(val.metadata["type"], module_name)
            if "type" in val.metadata
            else []
        )
